#!/usr/bin/env python
import sys, os, stat

# format
# header:
#   '!'+17bytes
# record (from 18th bytes...)
#   '+' prevpos arg ...
#   ':' continued
# end:
#   '.'
# other: err

__all__ = [ 'CreateTraDB', 'TraDBAppender', 'TraDBReader' ]


class LockError(ValueError): pass
class ParameterError(ValueError): pass
class InternalCheckingError(RuntimeError): pass


##  CreateTraDB
##
def CreateTraDB(fname, info, version='1'):
  if not isinstance(info, str) or ('\n' in info):
    raise ParameterError('Illegal value: %r' % info)
  fp = file(fname, 'w')
  fp.write('!T%c %08x %s\n' % (version, 0L, info))
  fp.write('.\n')
  fp.close()
  return


##  TraDBAppender
##
class TraDBAppender:

  def __init__(self, fname, info, version='1'):
    if not isinstance(info, str) or ('\n' in info):
      raise ParameterError('Illegal value: %r' % info)
    self.version = version
    self.fname = fname
    self.fname_locked = fname+'.locked'
    try:
      os.rename(self.fname, self.fname_locked)  # err
    except OSError, e:
      raise LockError('Cannot obtain lock: %s: %r' % (e, self.fname))
    try:
      size = os.stat(self.fname_locked)[stat.ST_SIZE] # err
      self.fp = file(self.fname_locked, 'r+')   # err
    except (OSError, IOError), e:
      try:
        os.rename(self.fname_locked, self.fname)  # err
      except OSError:
        pass
      raise LockError('Cannot open file: %s: %r' % (e, self.fname))
    #
    line = self.fp.readline()           # err or EOF
    if len(line) < 14 or not line.startswith('!T%c ' % self.version):
      raise InternalCheckingError('Illegal header: %r' % fname)
    pos0 = long(line[4:12], 16)         # ValueError
    if size <= pos0:
      raise InternalCheckingError('Too small file: %r' % fname)
    if not pos0:
      # empty db
      self.prevpos = self.fp.tell()
      line = self.fp.readline()
      if not line.startswith('.'):
        raise InternalCheckingError('Illegal character ("." expected): %r at %d' % (fname, self.prevpos))
    else:
      # seek to the end
      self.fp.seek(pos0)
      line = self.fp.readline()
      if not line.startswith('+'):
        raise InternalCheckingError('Illegal character ("+" expected): %r at %d' % (fname, pos0))
      while 1:
        self.prevpos = self.fp.tell()
        line = self.fp.readline()
        if not line or line[0] not in ':.':
          raise InternalCheckingError('Illegal character (":" or "." expected): %r at %d' % (fname, self.prevpos))
        if line[0] == '.': break
    self.fp.seek(self.prevpos)
    self.fp.write('+%08x %s\n' % (pos0, info))
    return
  
  def _writeline(self, line):
    if not isinstance(line, str) or line.find('\n') != len(line)-1:
      raise ParameterError('Text must contain only one newline character at the end: %r' % line)
    self.fp.write(':'+line)
    return
  
  def write(self, text):
    if not isinstance(text, str) or not text.endswith('\n'):
      raise ParameterError('Text must end with a newline: %r' % text)
    pos0 = 0
    while 1:
      try:
        pos1 = text.index('\n', pos0)+1
      except ValueError:
        break
      self._writeline(text[pos0:pos1])
      pos0 = pos1
    return

  def emclose(self):
    if self.fp:
      self.fp.close()
      print >>sys.stderr, 'emergency unlock: %r' % self.fname
      try:
        os.rename(self.fname, self.fname_locked)
      except OSError:
        pass
      self.fp = None
    return
  
  def close(self):
    self.fp.seek(0) # to begin
    self.fp.write('!T%c %08x' % (self.version, self.prevpos))
    self.fp.seek(0, 2) # to end
    self.fp.write('.\n')
    self.fp.close()
    os.rename(self.fname_locked, self.fname)
    self.fp = None
    return
    

##  TraRecordReader
##
class TraRecordReader:
  
  def __init__(self, fp, pos):
    self.fp = fp
    self.fp.seek(pos)
    line = self.fp.readline()
    if not line.startswith('+'):
      raise InternalCheckingError('Illegal character ("+" expected): %r' % fp)
    self.pos = self.fp.tell()
    self.info = line[10:].strip()
    return

  def __iter__(self):
    while 1:
      line = self.readline()
      if not line: break
      yield line
    return

  def readline(self):
    self.fp.seek(self.pos)
    line = self.fp.readline()
    if len(line) < 2 or line[0] not in ':+.':
      raise InternalCheckingError('Illegal character (":", "+" or "." expected): %r at %d' % (self.fp, self.pos))
    if line[0] == ':':
      self.pos = self.fp.tell()
      return line[1:]
    else:
      return ''

  def readlines(self):
    while 1:
      line = self.readline()
      if not line: break
      yield line
    return

  def read(self):
    r = []
    while 1:
      line = self.readline()
      if not line: break
      r.append(line)
    return ''.join(r)

  def close(self):
    self.fp = None
    return

##  TraDBReader
##
class TraDBReader:

  def __init__(self, fname, version='1'):
    self.version = version
    self.fname = fname
    try:
      self.fp = file(self.fname, 'r')     # err
    except IOError, e:
      raise LockError('Cannot open file: %s: %r' % (e, self.fname))
    line = self.fp.readline()           # err or EOF
    if len(line) < 14 or not line.startswith('!T%c ' % self.version):
      raise InternalCheckingError('Illegal header: %r' % fname)
    self.curpos = long(line[4:12], 16)  # ValueError
    if not self.curpos:
      raise InternalCheckingError('Empty record: %r' % fname)
    self.info = line[14:].strip()
    self.firstpos = self.fp.tell()
    self.lastpos = self.curpos
    self._read_header()
    return

  def get_db_info(self):
    return self.info

  def close(self):
    self.fp.close()
    return

  def _read_header(self):
    self.fp.seek(self.curpos)           # err
    line = self.fp.readline()           # err
    if not line.startswith('+'):
      raise InternalCheckingError('Illegal character ("+" expected): %r at %d' % (self.fname, self.curpos))
    self.prevpos = long(line[1:9], 16)  # ValueError
    self.recinfo = line[10:].strip()
    return

  def get_record(self):
    return TraRecordReader(self.fp, self.curpos)

  def get_record_info(self):
    return self.recinfo

  def first(self):
    self.curpos = self.firstpos
    self._read_header()                 # err
    return
    
  def last(self):
    self.curpos = self.lastpos
    self._read_header()                 # err
    return
    
  def prev(self):
    if not self.prevpos:
      return False
    self.curpos = self.prevpos
    self._read_header()                 # err
    return True
  
  def next(self):
    self.fp.seek(self.curpos)           # err
    line = self.fp.readline()           # err
    if not line.startswith('+'):
      raise InternalCheckingError('Illegal character ("+" expected): %r at %d' % (self.fname, self.curpos))
    while 1:
      pos = self.fp.tell()
      line = self.fp.readline()           # err
      if len(line) < 2 or line[0] not in ':+.':
        raise InternalCheckingError('Illegal character (":", "+" or "." expected): %r at %d' % (self.fname, pos))
      if line[0] == '.': return False
      if line[0] == '+': break
    self.curpos = pos
    self._read_header()                 # err
    return True



# testy
def test(fname='./test.db', n=5):
  # create db
  CreateTraDB(fname, 'testa')
  # write test
  for i in range(n):
    if i % 2:
      fp = TraDBAppender(fname, 'record%d' % i)
      fp.write('foo%d\n' % i)
      fp.write('baa\nbaag\n')
    else:
      fp = TraDBAppender(fname, 'empty-record')
    fp.close()
  # read test
  db = TraDBReader(fname)
  i = n
  while 1:
    fp = db.get_record()
    i -= 1
    print 'Record %d: info=%r' % (i, fp.info), list(fp.readlines())
    if not db.prev(): break
  assert i == 0
  while 1:
    fp = db.get_record()
    print 'Record %d: info=%r' % (i, fp.info), list(fp.readlines())
    i += 1
    if not db.next(): break
  return
    
if __name__ == "__main__": test()
