#!/usr/bin/env python
#
# htmlutils.py
#
#  Copyright (c) 2005  Yusuke Shinyama <yusuke at cs dot nyu dot edu>
#  
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation
#  files (the "Software"), to deal in the Software without
#  restriction, including without limitation the rights to use,
#  copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following
#  conditions:
#  
#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.
#  
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
#  KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
#  WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
#  PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
#  COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import re
from urllib import quote, unquote
from htmlentitydefs import codepoint2name


DEFAULT_CHARSET = 'iso-8859-1'


# english, cyrillic, greek, japanese, chinese
ALPHABET = re.compile(ur'[a-zA-Z0-9\u0390-\u03cf\u0400-\u045f\u3040-\u30ff\u4e00-\u9fff\uff10-\uff19\uff20-\uff9f]')
def get_alphabets(s):
  return ALPHABET.findall(x)

SPACE = re.compile(ur'^\s+$')
def is_space(s):
  return SPACE.match(s)

REMOVE_SPACES = re.compile(ur'\s+')
def remove_spaces(s):
  return REMOVE_SPACES.sub(u' ', s)

def concat(r):
  return remove_spaces(u''.join(r)).strip()


# encode_element
KEY_ATTRS = dict.fromkeys('class align valign rowspan colspan'.split(' '))
def encode_element(e):
  keys = [ k for k in e.attrs.keys() if k in KEY_ATTRS ]
  keys.sort()
  return str(e.tag + ''.join([ ':%s=%s' % (k.lower(), e.attrs[k].lower()) for k in keys ]))


# html tags

BLOCK = dict.fromkeys(
  "head title body p h1 h2 h3 h4 h5 h6 tr td th dt dd li "
  "ul ol dir menu pre dl div center frameset "
  "blockquote table fieldset address".split(" "))

CDATA = dict.fromkeys('script style'.split(' '))

INLINE = dict.fromkeys(
  "! tt i b u s strike big small nobr em strong dfn code samp kbd var cite abbr "
  "acronym a applet object button font map q sub sup span bdo layer ilayer iframe "
  "select textarea label button".split(" "))

INLINE_IMMED = dict.fromkeys(
  "basefont br area link img param hr input option "
  "colgroup col frame isindex meta base embed".split(" "))

NON_NESTED = dict.fromkeys("form noembed noframes noscript nolayer".split(" "))

VALID_TAGS = {}
for d in (BLOCK, CDATA, INLINE, INLINE_IMMED, NON_NESTED):
  for t in d.iterkeys():
    assert t not in VALID_TAGS, t
    VALID_TAGS[t] = 1
    

# mappings from html charsets to python codecs.
CODECS = {
  "ascii": "us-ascii",
  "us-ascii": "us-ascii",
  "iso-8859-1": "iso-8859-1",
  "jis": "iso-2022-jp",
  "iso-2022-jp": "iso-2022-jp",
  "euc": "euc-jp",
  "euc-jp": "euc-jp",
  "euc_jp": "euc-jp",
  "x-euc-jp": "euc-jp",
  "sjis": "ms932",
  "x-sjis": "ms932",
  "shift_jis": "ms932",
  "shift-jis": "ms932",
  "utf": "utf-8",
  "utf8": "utf-8",
  "utf-8": "utf-8",
}

# getcodec
def getcodec(charset, default=DEFAULT_CHARSET):
  codec = CODECS[default]
  if charset and charset.lower() in CODECS:
    codec = CODECS[charset.lower()]
  return codec

def getencoder(charset, default=DEFAULT_CHARSET):
  codec = CODECS[default]
  if charset and charset.lower() in CODECS:
    codec = CODECS[charset.lower()]
  return (lambda x: x.encode(codec, 'replace'))

def getdecoder(charset, default=DEFAULT_CHARSET):
  codec = CODECS[default]
  if charset and charset.lower() in CODECS:
    codec = CODECS[charset.lower()]
  return (lambda x: unicode(x, codec, 'replace'))
  

QUOTEREFS = { '&':'&amp;', '<':'&lt;', '>':'&gt;', '"':'&quot;' }
# quotestr(s, codec)
def quotestr(s, codec="ascii", quote_special=True):
  """Convert special characters in the string s into proper entitierefs
  with the given codec. The return value must be raw (non-Unicode) string."""
  def quote1(c):
    if quote_special and c in QUOTEREFS:
      return QUOTEREFS[c]
    else:
      return c.encode(codec, 'xmlcharrefreplace')
  return str(''.join([ quote1(c) for c in s ]))

# urldecode(s)
ARGPAT = re.compile(r'([^=]+)=(.*)')
def urldecode(s):
  """Convert a query string s into a mapping type.
  Duplicated arguments are not supported.
  Each parameter must occur at most only once."""
  args = {}
  for arg1 in s.split("&"):
    m = ARGPAT.match(arg1)
    if m:
      args[unquote(m.group(1))] = unquote(m.group(2))
  return args

# attr2str
def attr2str(attrs):
  r = []
  for (k,v) in attrs.iteritems():
    if v == None:
      r.append(' '+quotestr(k))
    else:
      r.append(' %s="%s"' % (quotestr(k), quotestr(v)))
  return ''.join(r)
