#!/usr/bin/env python
# -*- encoding: euc-jp -*-
import sys, os, re, md5, stat
from xml.sax import handler, parseString
from difflib import SequenceMatcher
from db import CreateTraDB, TraDBAppender, TraDBReader, LockError
from utils import *


class TransactionError(RuntimeError): pass

ITEMIDPAT = re.compile(r'^[0-9a-f]{32}$')
INFOPAT = re.compile(r'(.)(\d+):(\d+):(.*)')
def getentinfo(db):
  m = INFOPAT.match(db.get_record_info())
  return (m.group(1), int(m.group(2)), int(m.group(3)), decode_db(m.group(4)))
  

##  ItemParser
##
class ItemParser(handler.ContentHandler):
  
  def __init__(self):
    self.dic = {}
    self.name = None
    return handler.ContentHandler.__init__(self)
  
  def startElement(self, name, attrs):
    assert name in ('item','title','link','description','m')
    if name == 'item':
      return
    elif name in ('title','link','description'):
      self.name = name
      self.t = 0
      self.chars = []
    else:
      attrs = dict(attrs)
      assert 't' in attrs and attrs['t'].isdigit()
      self.t = int(attrs['t'])
    return

  def endElement(self, name):
    assert name in ('item','title','link','description','m')
    if name in ('item','m'):
      return
    elif name in ('title','link','description'):
      assert self.name == name
      self.dic[name] = self.chars
      self.name = None
    return
  
  def characters(self, s):
    if self.name in ('title','link','description'):
      self.chars.extend([ (self.t, c) for c in s ])
    return

  def parse(self, s):
    parseString('<?xml version="1.0"?>'+s, self)
    self.link = ''.join([ c for (t,c) in self.dic.get('link', []) ]).strip()
    self.title = self.dic.get('title', [])
    self.description = self.dic.get('description', [])
    return self


##  ItemEntry
##
class ItemEntry:

  def __init__(self, item,
               revision=0, modtime=0, flag='N', link='',
               title=[], description=[], editor=''):
    self.item = item
    self.revision = revision
    self.modtime = modtime
    self.flag = flag
    self.link = link
    self.title = title
    self.description = description
    self.editor = editor
    return
  
  def __repr__(self):
    return '<ItemEntry: modtime=%d, editor=%s, link=%s>' % (self.modtime, self.editor, self.link)

  def dup(self):
    ent = ItemEntry(self.item)
    ent.revision = self.revision
    ent.modtime = self.modtime
    ent.flag = self.flag
    ent.link = self.link
    ent.title = self.title[:]
    ent.description = self.description[:]
    ent.editor = self.editor
    return ent

  def load(self, db):
    (self.flag, self.revision, self.modtime, self.editor) = getentinfo(db)
    fp = db.get_record()
    parser = ItemParser().parse(decode_db(fp.read()))
    fp.close()
    self.link = parser.link
    self.title = parser.title
    self.description = parser.description
    return

  def write(self, fname):
    info = '%s%d:%d:%s' % (self.flag, self.revision, self.modtime, self.editor)
    fp = TraDBAppender(fname, dbstr(info))
    fp.write(self.dump_xml()+'\n')
    fp.close()
    return

  def edit(self, s0, s1, cur):
    self.modtime = cur
    m = SequenceMatcher(None, [ c for (t,c) in s0 ], list(s1))
    p2a = 0
    seq = []
    for (p1,p2,n) in m.get_matching_blocks():
      if p2a < p2:
        seq.extend([ (cur,c) for c in s1[p2a:p2] ])
      p2a = p2+n
      seq.extend(s0[p1:p1+n])
    return seq
  def edit_title(self, s1, cur):
    self.title = self.edit(self.title, s1, cur)
    return
  def edit_description(self, s1, cur):
    self.description = self.edit(self.description, s1, cur)
    return

  def dump_xml(self):
    t0 = 0
    seq = [ '<item>' ]
    if self.link:
      seq.append('<link>%s</link>' % dbquote(self.link))
    for (k,s) in (('title',self.title), ('description',self.description)):
      r = []
      for (t,c) in s:
        if t != t0:
          r.append('<m t="%d"/>' % t)
        r.append(dbquote(c))
        t0 = t
      seq.append( '<%s>%s</%s>' % (k, ''.join(r), k) )
    seq.append('</item>')
    return ''.join(seq)


##  Item
##
class Item:

  def __init__(self, repository, itemid):
    self.repository = repository
    self.itemid = itemid
    self.fname = repository.getfilename(itemid)
    self.ent0 = None
    self.ent1 = None
    return

  def __repr__(self):
    return '<Item: %s>' % (self.itemid)

  def set_flag(self, flag):
    mode = os.stat(self.fname)[stat.ST_MODE] & 0777
    if flag:
      mode |= 0111
    else:
      mode &= 0666
    os.chmod(self.fname, mode)
    return

  def get_flag(self):
    return bool(os.stat(self.fname)[stat.ST_MODE] & 0111)

  def init(self, modtime, title, link, description):
    try:
      os.stat(self.fname)
      raise TransactionError(self.itemid)
    except OSError:
      CreateTraDB(self.fname, self.itemid)
      def f(s): return [ (0,c) for c in s ]
      ent = ItemEntry(self, modtime=modtime, link=link,
                      title=f(title), description=f(description))
      ent.write(self.fname)
    self.repository.update_recent(self.itemid)
    return

  def commit(self, title, description, editor, flag, cur):
    assert self.ent1
    self.ent1 = self.ent1.dup()
    self.ent1.modtime = cur
    self.ent1.flag = flag
    self.ent1.editor = editor
    self.ent1.edit_title(title, cur)
    self.ent1.edit_description(description, cur)
    self.ent1.revision += 1
    try:
      self.ent1.write(self.fname)
      self.repository.update_recent(self.itemid)
    except LockError, e:
      raise TransactionError(e)
    return self.ent1
    
  def getent(self, revision=-1, nhist=5):
    try:
      db = TraDBReader(self.fname)
    except LockError, e:
      raise TransactionError(e)
    self.ent1 = ItemEntry(self)
    self.ent1.load(db)
    if revision == -1:
      revision = self.ent1.revision
    history = []
    ent = None
    while db.prev():
      (flag, rev1, modtime, editor) = getentinfo(db)
      history.append((rev1, modtime, editor))
      if rev1 == revision:
        ent = ItemEntry(self)
        ent.load(db)
      if rev1 <= revision-nhist: break
    db.first()
    self.ent0 = ItemEntry(self)
    self.ent0.load(db)
    db.close()
    return (self.ent0, self.ent1, ent, history)


##  Repository
##
class Repository:

  def __init__(self, name, dir=''):
    self.name = name
    self.dir = dir or name
    return

  def get_info(self):
    fp = file('%s/info' % self.dir)
    parser = ItemParser().parse(decode_db(fp.read()))
    fp.close()
    def s(x): return ''.join([ c for (t,c) in x ])
    return (remsp(s(parser.title)), remsp(parser.link), remsp(s(parser.description)))

  def getfilename(self, itemid):
    return '%s/%s/%s.db' % (self.dir, itemid[:2], itemid)

  def update_recent(self, itemid):
    try:
      orig = '../%s/%s.db' % (itemid[:2], itemid)
      link = '%s/recent/%s' % (self.dir, itemid)
      os.symlink(orig, link)
    except OSError:
      pass
    return
  
  def new_item(self, modtime, title, link, description):
    itemid = md5.new(link.encode('utf-8')).hexdigest()
    item = Item(self, itemid)
    item.init(modtime, title, link, description)
    return item

  def get_item(self, itemid):
    return Item(self, itemid)

  def del_item(self, itemid):
    os.unlink('%s/recent/%s' % (self.dir, itemid))
    return

  def recent_items(self):
    ids = []
    for itemid in os.listdir('%s/recent/' % self.dir):
      if not ITEMIDPAT.match(itemid): continue
      try:
        (ent0, ent1, ent2, history) = Item(self, itemid).getent(nhist=0)
        ids.append((itemid, ent0, ent1))
      except TransactionError, e:
        pass
    return ids
