#!/usr/bin/env python
# -*- encoding: euc-jp -*-
import sys, os, re, time
from repository import Item, Repository, TransactionError
from utils import *

stderr = sys.stderr
lowerbound = max


##  ArchiveDB
##
class ArchiveDB:

  ITEMS_PER_PAGE = 100
  ARCHIVE_URL = 'http://traiss.tabesugi.net:8080/'
  ARCHIVE_PAT = re.compile(r'archive-(\d+)\.html')
  ITEMURL_PAT = re.compile(r'<li>\s*([^<]+)<a href="http://traiss.tabesugi.net:8080/[^/]+/([0-9a-f]+)[^"]+">([^<]+)</a>')
  
  def __init__(self, dirname):
    self.dirname = dirname
    self.lastfile = None
    self.lastindex = -1
    self.lastitemids = []
    self.lastitems = {}
    for fname in os.listdir(dirname):
      m = self.ARCHIVE_PAT.match(fname)
      if m:
        n = int(m.group(1))
        if self.lastindex < n:
          self.lastindex = n
          self.lastfile = fname
    if self.lastfile:
      #print >>stderr, 'latest archive file: %r' % self.lastfile
      fp = file(dirname+'/'+self.lastfile)
      for line in fp:
        m = self.ITEMURL_PAT.search(line)
        if m:
          (date,itemid,title) = m.groups()
          self.lastitemids.append(itemid)
          title = remsp(unicode(title, HTML_ENCODING, 'replace'))
          date = remsp(unicode(date, HTML_ENCODING, 'replace'))
          self.lastitems[itemid] = (title, date)
      fp.close()
    return

  def archive_file(self, n):
    return 'archive-%05d.html' % n

  def dump_items(self, n, fname, triples):
    def elimtag(s):
      return s.encode(HTML_ENCODING, 'xmlcharrefreplace').replace('<','&lt;').replace('>','&gt;')
    #print >>stderr, 'write file: %r' % fname
    fp = file('%s/%s' % (self.dirname, fname), 'w')
    pagetitle = '%s/%s' % (repository.name, fname)
    print >>fp, '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"><html><head>'
    print >>fp, '<meta http-equiv="Content-Type" content="text/html; charset=%s">' % HTML_ENCODING
    print >>fp, '<title>%s</title>' % htmlquote(pagetitle)
    print >>fp, '</head><body><h1>%s</h1>' % htmlquote(pagetitle)
    if 0 <= n:
      fname = self.archive_file(n)
      print >>fp, '<p><a href="%s">[Older Articles]</a>' % (fname, )
    print >>fp, '<hr><ul>'
    triples.sort(lambda (i1,(t1,d1)),(i2,(t2,d2)): cmp(d2,d1))
    for (itemid,(title,date)) in triples:
      print >>fp, '<li> %s <a href="%s%s/%s?edit=0">%s</a>' % (htmlquote(date), self.ARCHIVE_URL, repository.name,
                                                               itemid, elimtag(title))
    print >>fp, '</ul><hr></body></html>'
    fp.close()
    return

  def add1(self, itemid, title, date):
    # assert item(itemid).get_flag() == True
    if itemid not in self.lastitems:
      self.lastitems[itemid] = (title, date)
      self.lastitemids.append(itemid)
      #print >>stderr, 'add1: %r' % itemid
    return

  def finish(self):
    n = lowerbound(0, self.lastindex)
    i = len(self.lastitemids)
    while 0 < i:
      fname = self.archive_file(n)
      ids = self.lastitemids[lowerbound(0, i-self.ITEMS_PER_PAGE):i]
      self.dump_items(n-1, fname, [ (id1, self.lastitems[id1]) for id1 in ids ])
      self.lastindex = n
      n += 1
      i -= self.ITEMS_PER_PAGE
    return

  def make_current(self, repository, itemids, force, timeout):
    items = []
    for itemid in itemids:
      try:
        (ent0, ent1, ent2, history) = Item(repository, itemid).getent(nhist=0)
        items.append((itemid, ent0, ent1))
      except TransactionError, e:
        pass
    items.sort(lambda (ida,a0,a1),(idb,b0,b1): cmp(b0.modtime, a0.modtime))
    triples = []
    deleted = []
    time0 = int(time.time())-timeout
    # new -> old
    for (itemid, ent0, ent1) in items:
      title = remsp(dump_raw(ent1.title))
      if ent1.flag != 'P':
        title = u'[̤] %s' % title
      date = time.strftime('%Y/%m/%d %H:%M', time.localtime(ent0.modtime))
      item = repository.get_item(itemid)
      if ent0.modtime < time0 and (force or not item.get_flag()):
        self.add1(itemid, title, date)
        item.set_flag(True)
      if ent1.modtime < time0:
        deleted.append(itemid)
      else:
        triples.append( (itemid, (title, date)) )
    self.finish()
    self.dump_items(self.lastindex, 'index.html', triples)
    return deleted



# main
ITEMIDPAT = re.compile(r'([0-9a-f]{32})')
if __name__ == '__main__':
  import getopt
  def usage():
    print "usage: archive.py [-i] [-f] [-d] [-t sec] archivedir repname [repdirname]"
    sys.exit(2)
  try:
    (opts, args) = getopt.getopt(sys.argv[1:], "ifdt:")
  except getopt.GetoptError:
    usage()
  (usestdin, force, delete, timeout) = (False, False, False, 86400*3)
  for (k,v) in opts:
    if k == '-i': usestdin = True
    elif k == '-f': force = True
    elif k == '-d': delete = True
    elif k == '-t': timeout = int(v)
  if len(args) == 2:
    (archivedir, repname) = args
    dirname = ''
  elif len(args) == 3:
    (archivedir, repname, repdirname) = args
  else:
    usage()
  #
  repository = Repository(repname, repdirname)
  if usestdin:
    itemids = []
    for line in sys.stdin:
      m = ITEMIDPAT.search(line.strip())
      if m:
        itemids.append(m.group(1))
  else:
    itemids = [ itemid for itemid in os.listdir('%s/recent/' % repository.dir)
                if ITEMIDPAT.match(itemid) ]
  db = ArchiveDB(archivedir)
  deleted = db.make_current(repository, itemids, force, timeout)
  if delete and deleted:
    for itemid in deleted:
      #print >>stderr, 'deleted recent: %r' % itemid
      fname = '%s/recent/%s' % (repository.dir, itemid)
      os.unlink(fname)
