#!/usr/bin/env python
# -*- encoding: euc-jp -*-

import sys, re, urllib
from difflib import SequenceMatcher
from htmlparser3 import HTMLHandler, HTMLParser3
from htmlutils import quotestr, attr2str

__all__ = [
  'TRAISS_TOPURL', 'TRAISS_VERSION', 'HTML_ENCODING', 'DB_ENCODING', 'RSS_ENCODING',
  'htmlquote', 'dbquote', 'dbstr', 'rssquote', 'urlquote',
  'safeint', 'decode_form', 'decode_db', 'remsp', 'repr_time',
  'diff', 'dump_raw', 'dump_style', 'riptags', 'ripalltags'
  ]

TRAISS_TOPURL = 'http://traiss.tabesugi.net/'
TRAISS_VERSION='0.22'

HTML_ENCODING = 'euc-jp'
DB_ENCODING = 'euc-jp'
RSS_ENCODING = 'utf-8'

def htmlquote(s):
  return quotestr(s, HTML_ENCODING)

def dbquote(s):
  return quotestr(s, DB_ENCODING)
def dbstr(s):
  return s.encode(DB_ENCODING, 'replace')

def rssquote(s):
  return quotestr(s, RSS_ENCODING)

def urlquote(s):
  return urllib.quote(s.encode(HTML_ENCODING, 'replace'))

def safeint(x, v=0):
  try:
    return int(x)
  except (TypeError, ValueError):
    return v

SANITIZE_PAT = re.compile(ur'[\n\u0020-\uffff]+')
def decode_form(form, k, limit=10000):
  s = form.getvalue(k, '')
  try:
    s = unicode(s, HTML_ENCODING, 'replace')
  except UnicodeError:
    return u''
  return u''.join(SANITIZE_PAT.findall(s)).strip()[:limit]

def decode_db(s):
  return unicode(s, DB_ENCODING, 'replace')

REMSP = re.compile(r'\s+')
def remsp(s):
  return REMSP.sub(' ', s.strip())


def styleid(t, cur, highlight=0):
  if t == -1:
    return 'd'                          # diff
  elif highlight and t == highlight:
    return 's0'
  elif t == 0:
    return 's4'
  else:
    d = cur - t
    if d < 3600:                        # 1hour
      return 's1'
    elif d < 21600:                     # 6hour
      return 's2'
    elif d < 86400:                     # 1day
      return 's3'
  return 4

def repr_time(t, cur):
  d = cur-t
  if d < 3600:
    s = u'%dmins' % (d/60)
  elif d < 86400:
    s = u'%dhours' % ((d+1800)/3600)
  else:
    s = '%ddays' % ((d+43200)/86400)
  return '<span class="%s">%s</span>' % (styleid(t, cur), s)

def diff(s1, s2):
  m = SequenceMatcher(None, s1, s2)
  (r1, r2) = ([], [])
  (p1a, p2a) = (0, 0)
  for (p1,p2,n) in m.get_matching_blocks():
    if p1a < p1:
      r1.extend([ (-1,c) for (t,c) in s1[p1a:p1] ])
    r1.extend([ (t,c) for (t,c) in s1[p1:p1+n] ])
    p1a = p1 + n
    if p2a < p2:
      r2.extend([ (-1,c) for (t,c) in s2[p2a:p2] ])
    r2.extend([ (0,c) for (t,c) in s2[p2:p2+n] ])
    p2a = p2 + n
  return (r1, r2)

def dump_raw(s):
  return ''.join([ c for (t,c) in s ])

def dump_style(s, cur, highlight=0):
  r = []
  style0 = None
  for (t,c) in s:
    style1 = styleid(t, cur, highlight)
    if style0 != style1:
      if style0 != None:
        r.append('</span>')
      r.append('<span class="%s">' % style1)
      style0 = style1
    if c == '\n':
      r.append('<br>')
    else:
      r.append(htmlquote(c))
  if style0 != -1:
    r.append('</span>')
  return ''.join(r)


##  TagRipper
##
class TagRipper(HTMLHandler):

  ESCAPE = { '&':'&amp;', '<':'&lt;', '>':'&gt;' }
  ALLOW_TAGS = {
    'br': (),
    'a': ('href', 'style'),
    'b': ('style',),
    'i': ('style',),
    'u': ('style',),
    's': ('style',),
    'q': ('style',),
    'span': ('style',),
    'strong': ('style',),
    'em': ('style',),
    'code': ('style',),
    'kbd': ('style',),
    'small': ('style',),
    'big': ('style',),
    }

  def __init__(self, allow_tags=ALLOW_TAGS):
    self.buf = []
    self.tags = []
    self.allow_tags = allow_tags
    return

  def handle_data(self, s):
    self.buf.append(''.join( [ self.ESCAPE.get(c, c) for c in s ]))
    return

  def start_unknown(self, tag, attrs):
    try:
      allow_fields = self.allow_tags[tag]
      for k in attrs.keys():
        if k not in allow_fields:
          del attrs[k]
      self.buf.append('<%s%s>' % (tag, attr2str(attrs)))
      if tag != 'br':
        self.tags.append(tag)
    except KeyError:
      return

  def end_unknown(self, tag):
    if self.tags and self.tags[-1] == tag:
      self.tags.pop()
      self.buf.append('</%s>' % tag)
    return

  def finish(self):
    while self.tags:
      tag = self.tags.pop()
      self.buf.append('</%s>' % tag)
    return ''.join(self.buf)

def riptags(s):
  return HTMLParser3(TagRipper()).feed_unicode(s).close()

def ripalltags(s):
  return HTMLParser3(TagRipper(allow_tags={})).feed_unicode(s).close()
